from django.contrib.auth.models import AbstractBaseUser
from django.db import models
from django.utils.translation import ugettext_lazy as _

from chamber import models as chamber_models
from chamber.models import fields as chamber_fields
from chamber.models.dispatchers import CreatedDispatcher, PropertyDispatcher, StateDispatcher
from chamber.models.signals import dispatcher_pre_save, dispatcher_post_save
from chamber.utils.datastructures import ChoicesNumEnum, SequenceChoicesNumEnum, SubstatesChoicesNumEnum

from .handlers import (create_test_dispatchers_model_handler, create_test_fields_model_handler,
                       create_test_smart_model_handler, create_csv_record_handler)


class ShortcutsModel(models.Model):
    name = models.CharField(max_length=100)
    datetime = models.DateTimeField()
    number = models.IntegerField()


class DiffModel(chamber_models.SmartModel):
    name = models.CharField(max_length=100)
    datetime = models.DateTimeField()
    number = models.IntegerField()


class ComparableModel(chamber_models.ComparableModelMixin, models.Model):
    name = models.CharField(max_length=100)


class TestSmartModel(chamber_models.SmartModel):
    name = models.CharField(max_length=100)


class RelatedSmartModel(chamber_models.SmartModel):
    test_smart_model = models.ForeignKey(TestSmartModel, related_name='test_smart_models', on_delete=models.CASCADE)


class BackendUser(AbstractBaseUser):
    pass


class FrontendUser(AbstractBaseUser):
    pass


class CSVRecord(chamber_models.SmartModel):
    name = models.CharField(max_length=100, null=True, blank=True)
    number = models.IntegerField(null=True, blank=True)


class TestFieldsModel(chamber_models.SmartModel):

    STATE = ChoicesNumEnum(
        ('OK', _('ok'), 1),
        ('NOT_OK', _('not ok'), 2),
    )

    STATE_REASON = SubstatesChoicesNumEnum({
        STATE.OK: (
            ('SUB_OK_1', _('1st substate'), 1),
            ('SUB_OK_2', _('2nd substate'), 2),
        ),
        STATE.NOT_OK: (
            ('SUB_NOT_OK_1', _('1st not ok substate'), 3),
            ('SUB_NOT_OK_2', _('2nd not ok substate'), 4),
        ),
    })

    GRAPH = SequenceChoicesNumEnum((
        ('FIRST', _('first'), 1, ('SECOND',)),
        ('SECOND', _('second'), 2, ('THIRD',)),
        ('THIRD', _('third'), 3, ()),
    ), initial_states=('FIRST',))

    decimal = chamber_fields.DecimalField(null=True, blank=True, min=3, max=10, max_digits=5, decimal_places=3)
    state = models.IntegerField(null=True, blank=False, choices=STATE.choices, default=STATE.OK)
    state_reason = chamber_models.SubchoicesPositiveIntegerField(null=True, blank=True, enum=STATE_REASON,
                                                                 supchoices_field_name='state',
                                                                 default=STATE_REASON.SUB_OK_1)
    state_prev = chamber_models.PrevValuePositiveIntegerField(verbose_name=_('previous state'), null=False, blank=False,
                                                              copy_field_name='state', choices=STATE.choices,
                                                              default=STATE.NOT_OK)
    state_graph = chamber_models.EnumSequencePositiveIntegerField(verbose_name=_('graph'), null=True, blank=True,
                                                                  enum=GRAPH)
    file = chamber_models.FileField(verbose_name=_('file'), null=True, blank=True,
                                    allowed_content_types=('application/pdf', 'text/plain', 'text/csv'))
    image = chamber_models.ImageField(verbose_name=_('image'), null=True, blank=True, max_upload_size=1)
    price = chamber_models.PriceField(verbose_name=_('price'), null=True, blank=True, currency=_('EUR'))
    total_price = chamber_models.PositivePriceField(verbose_name=_('total price'), null=True, blank=True)


class TestDispatchersModel(chamber_models.SmartModel):

    STATE = ChoicesNumEnum(
        ('FIRST', _('first'), 1),
        ('SECOND', _('second'), 2),
    )
    state = models.IntegerField(null=True, blank=False, choices=STATE.choices, default=STATE.FIRST)

    dispatchers = (
        CreatedDispatcher(create_csv_record_handler, signal=dispatcher_pre_save),
        StateDispatcher(create_test_smart_model_handler, STATE, state, STATE.SECOND, signal=dispatcher_pre_save),
        PropertyDispatcher(create_test_fields_model_handler, 'always_dispatch', signal=dispatcher_post_save),
        PropertyDispatcher(create_test_dispatchers_model_handler, 'never_dispatch', signal=dispatcher_post_save),
    )

    @property
    def always_dispatch(self):
        return True

    @property
    def never_dispatch(self):
        return False



from dj.settings.base import *  # pylint: disable=E0401


DEBUG = TEMPLATE_DEBUG = THUMBNAIL_DEBUG = True

SITE_ID = BACKEND_SITE_ID

ALLOWED_HOSTS = [DOMAINS.get(SITE_ID).hostname]

# URL with protocol (and port)
PROJECT_URL = DOMAINS.get(SITE_ID).url

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_DIR, 'var', 'db', 'sqlite.db'),
        'USER': '',
        'PASSWORD': '',
    },
}

ROOT_URLCONF = DOMAINS.get(SITE_ID).urlconf

STATIC_ROOT = ''

# Additional locations of static files
STATICFILES_DIRS = (
    STATICFILES_ROOT,
)



from distutils.version import StrictVersion

import django

from django.conf.urls import url

from test_chamber import views  # pylint: disable=E0401


if StrictVersion(django.get_version()) < StrictVersion('1.9'):
    from django.conf.urls import patterns

    urlpatterns = patterns(
        '',
        url(r'^current_time_frontend/$', views.current_datetime, name='current-datetime')
    )
else:
    urlpatterns = [
        url(r'^current_time_frontend/$', views.current_datetime, name='current-datetime')
    ]



from django import forms
from django.utils.translation import ugettext

from chamber import config

from .validators import (
    RestrictedFileValidator, AllowedContentTypesByFilenameFileValidator, AllowedContentTypesByContentFileValidator
)

class DecimalField(forms.DecimalField):

    def __init__(self, *args, **kwargs):
        self.step = kwargs.pop('step', 'any')
        self.min = kwargs.pop('min', None)
        self.max = kwargs.pop('max', None)
        super().__init__(*args, **kwargs)

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget)
        attrs['step'] = self.step
        if self.min is not None:
            attrs['min'] = self.min
        if self.max is not None:
            attrs['max'] = self.max
        return attrs


class PriceNumberInput(forms.NumberInput):

    def __init__(self, currency, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.placeholder = currency


class PriceField(DecimalField):

    widget = PriceNumberInput

    def __init__(self, *args, **kwargs):
        currency = kwargs.pop('currency', ugettext('CZK'))
        kwargs.setdefault('max_digits', 10)
        kwargs.setdefault('decimal_places', 2)
        if 'widget' not in kwargs:
            kwargs['widget'] = PriceNumberInput(currency)
        super().__init__(*args, **kwargs)


class RestrictedFileField(forms.FileField):

    def __init__(self, *args, **kwargs):
        max_upload_size = kwargs.pop('max_upload_size', config.CHAMBER_MAX_FILE_UPLOAD_SIZE) * 1024 * 1024
        allowed_content_types = kwargs.pop('allowed_content_types', None)

        validators = tuple(kwargs.pop('validators', [])) + (
            RestrictedFileValidator(max_upload_size),
        )
        if allowed_content_types:
            validators += (
                AllowedContentTypesByFilenameFileValidator(allowed_content_types),
                AllowedContentTypesByContentFileValidator(allowed_content_types),
            )
        super().__init__(validators=validators, *args, **kwargs)



from django.utils import numberformat
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe


def natural_number_with_currency(number, currency, show_decimal_place=True, use_nbsp=True):
    """
    Return a given `number` formatter a price for humans.
    """
    humanized = '{} {}'.format(
        numberformat.format(
            number=number,
            decimal_sep=',',
            decimal_pos=2 if show_decimal_place else 0,
            grouping=3,
            thousand_sep=' ',
            force_grouping=True
        ),
        force_text(currency)
    )
    return mark_safe(humanized.replace(' ', '\u00a0')) if use_nbsp else humanized



from collections import OrderedDict

from django.http.request import QueryDict
from django.utils.encoding import force_text


def query_string_from_dict(qs_dict):
    qs_prepared_dict = OrderedDict()
    for key, val in qs_dict.items():
        if isinstance(val, list):
            val = '[%s]' % ','.join([force_text(v) for v in val])
        qs_prepared_dict[key] = val

    qdict = QueryDict('').copy()
    qdict.update(qs_prepared_dict)
    return qdict.urlencode()



from django.core.files.base import ContentFile
from django.test import TestCase

from chamber.storages.boto3 import force_bytes_content

from germanium.decorators import data_provider


class S3StorageTestCase(TestCase):

    TEST_DATA = (
        ('Hello, this is str content.', b'Hello, this is str content.', True),
        (b'Hello, this is bytes content.', b'Hello, this is bytes content.', False),
    )

    @data_provider(TEST_DATA)
    def test_s3storage_casts_content_to_bytes_if_needed(self, content, content_result, should_cast):
        file = ContentFile(content)
        result, casted = force_bytes_content(file)
        casted_content = result.read()
        self.assertEqual(should_cast, casted)
        self.assertEqual(bytes, type(casted_content))
        self.assertEqual(content_result, casted_content)



from nose.tools import raises  # pylint: disable=E0401

from django.core.exceptions import ImproperlyConfigured
from django.test import TransactionTestCase

from chamber.models.dispatchers import BaseDispatcher, StateDispatcher
from chamber.shortcuts import change_and_save

from germanium.tools import assert_equal  # pylint: disable=E0401

from test_chamber.models import (CSVRecord, TestDispatchersModel, TestFieldsModel,  # pylint: disable=E0401
                                 TestSmartModel)  # pylint: disable=E0401


class DispatchersTestCase(TransactionTestCase):

    def test_state_dispatcher(self):
        m = TestDispatchersModel.objects.create()

        # Moving TestDispatcher model to SECOND state should create new TestSmartModel instance
        assert_equal(TestSmartModel.objects.count(), 0)
        change_and_save(m, state=TestDispatchersModel.STATE.SECOND)
        assert_equal(TestSmartModel.objects.count(), 1)

        # But subsequent saves should not create more instances
        change_and_save(m, state=TestDispatchersModel.STATE.SECOND)
        assert_equal(TestSmartModel.objects.count(), 1)

        # Moving back and forth between the states creates another instance
        change_and_save(m, state=TestDispatchersModel.STATE.FIRST)
        change_and_save(m, state=TestDispatchersModel.STATE.SECOND)
        assert_equal(TestSmartModel.objects.count(), 2)

    def test_property_dispatcher(self):
        # Saving the model should always fire up the one property handler, not the second
        assert_equal(TestFieldsModel.objects.count(), 0)
        TestDispatchersModel.objects.create()
        assert_equal(TestFieldsModel.objects.count(), 1)
        assert_equal(TestDispatchersModel.objects.count(), 1)

    def test_created_dispatcher(self):
        assert_equal(CSVRecord.objects.count(), 0)
        m = TestDispatchersModel.objects.create()
        assert_equal(CSVRecord.objects.count(), 1)
        change_and_save(m, state=TestDispatchersModel.STATE.SECOND)
        assert_equal(CSVRecord.objects.count(), 1)

    @raises(ImproperlyConfigured)
    def test_should_call_dispatcher_with_invalid_value(self):
        model = TestDispatchersModel.objects.create()
        model.state = 3
        field = TestDispatchersModel._meta.get_field('state')  # pylint: disable=W0212
        StateDispatcher(lambda: False, TestDispatchersModel.STATE, field, model.state)(model)

    def _create_model_and_invalid_field(self):
        model = TestDispatchersModel.objects.create()
        model.state = TestDispatchersModel.STATE.SECOND
        return model, TestDispatchersModel._meta.get_field('state')  # pylint: disable=W0212

    @raises(ImproperlyConfigured)
    def test_should_not_call_dispatcher_with_invalid_handler(self):
        model, field = self._create_model_and_invalid_field()
        StateDispatcher(None, TestDispatchersModel.STATE, field, model.state)(model)

    @raises(NotImplementedError)
    def test_should_call_invalid_dispatcher(self):
        class InvalidDispatcher(BaseDispatcher):  # pylint: disable=W0223

            def _validate_init_params(self):
                pass

        InvalidDispatcher(lambda: False)(object())
        BaseDispatcher(lambda: False)



from django.test import TestCase

from chamber.models.humanized_helpers import price_humanized

from germanium.decorators import data_provider  # pylint: disable=E0401
from germanium.tools import assert_equal  # pylint: disable=E0401


class HumanizedHelpersTestCase(TestCase):

    numbers_with_currencies = (
        (100.00, 'CZK', '100,00\xa0CZK'),
        (1000000.00, 'EUR', '1\xa0000\xa0000,00\xa0EUR'),
        (1000000.00, None, '1\xa0000\xa0000,00\xa0CZK'),
        (None, None, '(None)'),
    )

    @data_provider(numbers_with_currencies)
    def test_should_return_price_humanized(self, number, currency, expected):
        assert_equal(expected, price_humanized(number, None, currency))



from copy import copy
from datetime import datetime, time, timedelta

from django.utils import timezone
from django.utils.timezone import utc


def range_for_today():
    return range_for_day(datetime.today())


def range_for_day(day):
    return (datetime.combine(day, time.min).replace(tzinfo=utc),
            datetime.combine(day, time.max).replace(tzinfo=utc))


def range_for_month_by_day(day):
    start = copy(day).replace(day=1)
    return (datetime.combine(start, time.min).replace(tzinfo=utc),
            datetime.combine(day, time.max).replace(tzinfo=utc))


def range_for_current_month():
    return range_for_month_by_day(datetime.today())


def range_for_last_24_hours(now=datetime.now(tz=utc)):
    return (now - timedelta(hours=24), now)


def make_naive(datetime_object):
    return (timezone.make_naive(datetime_object, timezone.get_default_timezone())
            if timezone.is_aware(datetime_object) else datetime_object)


def make_aware(datetime_object):
    return timezone.make_aware(datetime_object, timezone.get_default_timezone())


def date_to_datetime(date_object):
    return make_aware(datetime(year=date_object.year, month=date_object.month, day=date_object.day, hour=12))


def aware_datetime(*args, **kwargs):
    return make_aware(datetime(*args, **kwargs))



from django.core.files.uploadedfile import SimpleUploadedFile
from django.forms import TextInput, ValidationError
from django.test import TransactionTestCase

from germanium.tools import assert_equal, assert_true, assert_raises, assert_not_raises  # pylint: disable=E0401

from chamber.forms.fields import DecimalField, RestrictedFileField


__all__ = (
    'FormFieldsTestCase',
)


class FormFieldsTestCase(TransactionTestCase):

    def test_decimal_field_should_return_correct_widget_attrs(self):
        kwargs = {
            'step': 0.5,
            'min': 1.0,
            'max': 10.0,
        }
        field = DecimalField(**kwargs)
        widget_attrs = field.widget_attrs(TextInput())
        assert_true(len(widget_attrs.keys()) > 0)
        for attr, value in kwargs.items():
            assert_equal(value, widget_attrs[attr])

    def test_decimal_field_should_return_default_attrs(self):
        field = DecimalField()
        widget_attrs = field.widget_attrs(TextInput())
        assert_equal({'step': 'any'}, widget_attrs)

    def test_restricted_file_field_should_raise_validation_error_for_invalid_files(self):
        field = RestrictedFileField(allowed_content_types=('image/jpeg', 'application/pdf'), max_upload_size=1)

        # Invalid file type
        with open('data/all_fields_filled.csv', 'rb') as f:
            with assert_raises(ValidationError):
                field.clean(SimpleUploadedFile('all_fields_filled.pdf', f.read()))

        # Invalid size
        with open('data/test.jpg', 'rb') as f:
            with assert_raises(ValidationError):
                field.clean(SimpleUploadedFile('test.jpg', f.read()))

        # Invalid file suffix
        with open('data/test.pdf', 'rb') as f:
            with assert_raises(ValidationError):
                field.clean(SimpleUploadedFile('test.csv', f.read()))

        # Valid file
        with open('data/test.pdf', 'rb') as f:
            with assert_not_raises(ValidationError):
                field.clean(SimpleUploadedFile('test.pdf', f.read()))



from .formatters import *  # NOQA
from .forms import *  # NOQA
from .importers import *  # NOQA
from .models import *  # NOQA
from .multidomains import *  # NOQA
from .shortcuts import *  # NOQA
from .storages import *  # NOQA
from .utils import *  # NOQA



from chamber.commands import ImportCSVCommand

from test_chamber.models import CSVRecord  # pylint: disable=E0401


class Command(ImportCSVCommand):
    model_class = CSVRecord
    fields = ('id', 'name', 'number')
    csv_path = 'data/all_fields_filled.csv'

    def clean_number(self, value):
        # Just to test clean methods are called
        return 888



from django.test import TestCase
from django.utils.translation import get_language

from chamber.utils.decorators import classproperty, singleton, translation_activate_block

from germanium.tools import assert_equal  # pylint: disable=E0401


class ObjectWithClassProperty(object):

    @classproperty
    @classmethod
    def class_method_property(cls):
        return 1


@singleton
class SingletonObject(object):

    num_singletons = 0

    def __init__(self):
        self.num_singletons += 1
        super(self.__class__, self).__init__()

    def get_count_singletons(self):
        return self.num_singletons


class DecoratorsTestCase(TestCase):

    def test_class_property(self):
        assert_equal(ObjectWithClassProperty.class_method_property, 1)
        assert_equal(ObjectWithClassProperty().class_method_property, 1)

    def test_singleton(self):
        singleton = SingletonObject()
        assert_equal(singleton.num_singletons, 1)
        singleton = SingletonObject()
        assert_equal(singleton.num_singletons, 1)

    def test_translation_activate_block(self):

        @translation_activate_block(language='en')
        def en_block():
            assert_equal(get_language(), 'en')

        assert_equal(get_language(), 'cs')
        en_block()
        assert_equal(get_language(), 'cs')



def create_test_smart_model_handler(instance, **kwargs):
    from .models import TestSmartModel  # NOQA

    TestSmartModel.objects.create(name='name')


def create_test_fields_model_handler(instance, **kwargs):
    from .models import TestFieldsModel  # NOQA

    TestFieldsModel.objects.create()


def create_test_dispatchers_model_handler(instance, **kwargs):
    from .models import TestDispatchersModel  # NOQA

    TestDispatchersModel.objects.create()


def create_csv_record_handler(instance, **kwargs):
    from .models import CSVRecord  # NOQA

    CSVRecord.objects.create()



from django.conf import settings


CHAMBER_MAX_FILE_UPLOAD_SIZE = getattr(settings, 'CHAMBER_MAX_FILE_UPLOAD_SIZE', 20)
CHAMBER_MULTIDOMAINS_OVERTAKER_AUTH_COOKIE_NAME = getattr(settings, 'CHAMBER_MULTIDOMAINS_OVERTAKER_AUTH_COOKIE_NAME',
                                                          None)
CHAMBER_DEFAULT_IMAGE_ALLOWED_CONTENT_TYPES = getattr(settings, 'CHAMBER_DEFAULT_IMAGE_ALLOWED_CONTENT_TYPES',
                                                      {'image/jpeg', 'image/png', 'image/gif'})



from django.test import TestCase

from chamber.formatters import natural_number_with_currency

from germanium.decorators import data_provider  # pylint: disable=E0401
from germanium.tools import assert_equal  # pylint: disable=E0401


class FormattersTestCase(TestCase):

    numbers_with_currencies = (
        (100.00, 'CZK', False, False, '100 CZK'),
        (100.00, 'CZK', True, False, '100,00 CZK'),
        (100.00, 'CZK', False, True, '100\xa0CZK'),
        (1000000.00, 'CZK', False, True, '1\xa0000\xa0000\xa0CZK'),
    )

    @data_provider(numbers_with_currencies)
    def test_should_return_natural_number_with_currency(self, number, currency, show_decimal_place, use_nbsp, expected):
        assert_equal(expected, natural_number_with_currency(number, currency, show_decimal_place, use_nbsp))



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

from chamber.multidomains.domain import Domain, get_current_domain, get_domain, get_domain_choices, get_user_class
from chamber.multidomains.urlresolvers import reverse

from germanium.tools import assert_equal, assert_raises  # pylint: disable=E0401

from test_chamber.models import BackendUser, FrontendUser  # pylint: disable=E0401


class MultidomainsTestCase(TestCase):

    def test_get_domain_choices(self):
        choices = get_domain_choices()
        assert_equal(len(choices), 2)
        assert_equal(choices[0][1], 'backend')
        assert_equal(choices[1][1], 'frontend')
        assert_equal(choices[0][0], 1)
        assert_equal(choices[1][0], 2)

    def test_get_current_user_class(self):
        assert_equal(get_user_class(), BackendUser)

    def test_get_user_class(self):
        assert_equal(get_domain(settings.BACKEND_SITE_ID).user_class, BackendUser)
        assert_equal(get_domain(settings.FRONTEND_SITE_ID).user_class, FrontendUser)

    def test_get_current_domain(self):
        assert_equal(get_current_domain().name, 'backend')

    def test_get_domain(self):
        assert_equal(get_domain(settings.BACKEND_SITE_ID).name, 'backend')
        assert_equal(get_domain(settings.FRONTEND_SITE_ID).name, 'frontend')
        assert_raises(ImproperlyConfigured, get_domain, 3)

    def test_get_domain_url(self):
        assert_equal(get_domain(settings.BACKEND_SITE_ID).url, 'http://localhost:8000')
        assert_equal(get_domain(settings.FRONTEND_SITE_ID).url, 'https://localhost')

    def test_new_domain_port(self):
        assert_equal(Domain('testA', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='http', hostname='localhost').port, 80)
        assert_equal(Domain('testB', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='https', hostname='localhost').port, 443)
        assert_equal(Domain('testC', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='http', hostname='localhost', port=443).port, 443)
        assert_equal(Domain('testD', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='https', hostname='localhost', port=80).port, 80)
        assert_raises(ImproperlyConfigured, Domain, 'testF', 'dj.backend_urls', 'test_chamber.BackendUser',
                      protocol='hbbs', hostname='localhost')
        assert_equal(Domain('testD', 'dj.backend_urls', 'test_chamber.BackendUser',
                            url='https://localhost:80').port, 80)
        assert_equal(Domain('testD', 'dj.backend_urls', 'test_chamber.BackendUser',
                            url='https://localhost').port, 443)

        assert_equal(Domain('testA', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='http', hostname='localhost').url, 'http://localhost')
        assert_equal(Domain('testB', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='https', hostname='localhost').url, 'https://localhost')
        assert_equal(Domain('testC', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='http', hostname='localhost', port=443).url, 'http://localhost:443')
        assert_equal(Domain('testD', 'dj.backend_urls', 'test_chamber.BackendUser',
                            protocol='https', hostname='localhost', port=80).url, 'https://localhost:80')

    def test_reverse(self):
        assert_equal(reverse('current-datetime'), '/current_time_backend/')
        assert_equal(reverse('current-datetime', site_id=settings.BACKEND_SITE_ID), '/current_time_backend/')
        assert_equal(reverse('current-datetime', site_id=settings.FRONTEND_SITE_ID), '/current_time_frontend/')

        assert_equal(reverse('current-datetime', site_id=settings.BACKEND_SITE_ID, add_domain=True),
                     'http://localhost:8000/current_time_backend/')
        assert_equal(reverse('current-datetime', site_id=settings.FRONTEND_SITE_ID, add_domain=True),
                     'https://localhost/current_time_frontend/')

        assert_equal(reverse('current-datetime', site_id=settings.BACKEND_SITE_ID, add_domain=True,
                             urlconf='dj.frontend_urls'),
                     'http://localhost:8000/current_time_frontend/')
        assert_equal(reverse('current-datetime', site_id=settings.FRONTEND_SITE_ID, add_domain=True,
                             urlconf='dj.backend_urls'),
                     'https://localhost/current_time_backend/')

        assert_equal(reverse('current-datetime', qs_kwargs={'a': 1}), '/current_time_backend/?a=1')

        assert_raises(ImproperlyConfigured, reverse, 'current-datetime', site_id=3)



from django.test import TestCase

from chamber.utils.datastructures import ChoicesEnum, ChoicesNumEnum, Enum, NumEnum, OrderedSet

from germanium.tools import assert_equal, assert_raises, assert_is_none, assert_in, assert_not_in


class DatastructuresTestCase(TestCase):

    def test_enum_should_contain_only_defined_values(self):
        enum = Enum(
            'A', 'B'
        )
        assert_equal(enum.A, 'A')
        assert_equal(enum.B, 'B')
        assert_equal(list(enum), ['A', 'B'])
        assert_equal(enum.all, ('A', 'B'))
        assert_equal(enum.get_name('A'), 'A')
        with assert_raises(AttributeError):
            enum.C  # pylint: disable=W0104
        assert_is_none(enum.get_name('C'))
        assert_in('A', enum)
        assert_in(enum.A, enum)
        assert_not_in('C', enum)

    def test_enum_with_distinct_key_and_value_should_contain_only_defined_values(self):
        enum = Enum(
            ('A', 'c'), ('B', 'd')
        )
        assert_equal(enum.A, 'c')
        assert_equal(enum.B, 'd')
        assert_equal(list(enum), ['c', 'd'])
        assert_equal(enum.all, ('c', 'd'))
        assert_equal(enum.get_name('c'), 'A')
        with assert_raises(AttributeError):
            enum.C  # pylint: disable=W0104
        assert_is_none(enum.get_name('f'))
        assert_in('c', enum)
        assert_in(enum.A, enum)
        assert_not_in('A', enum)

    def test_auto_gemerated_num_enum_should_contain_only_defined_values(self):
        enum = NumEnum(
            'A', 'B'
        )
        assert_equal(enum.A, 1)
        assert_equal(enum.B, 2)
        assert_equal(list(enum), [1, 2])
        assert_equal(enum.all, (1, 2))
        assert_equal(enum.get_name(1), 'A')
        with assert_raises(AttributeError):
            enum.C  # pylint: disable=W0104
        assert_is_none(enum.get_name(3))
        assert_in(1, enum)
        assert_in(enum.A, enum)
        assert_not_in('A', enum)

    def test_num_enum_with_defined_ids_should_return_valid_values(self):
        enum = NumEnum(
            ('A', 4), ('B', 2), 'C'
        )
        assert_equal(enum.A, 4)
        assert_equal(enum.B, 2)
        assert_equal(enum.C, 3)
        assert_equal(list(enum), [4, 2, 3])

    def test_num_enum_same_numbers_should_raise_exception(self):
        with assert_raises(ValueError):
            NumEnum(
                ('A', 1), ('B', 1)
            )

    def test_num_enum_same_generated_numbers_should_raise_exception(self):
        with assert_raises(ValueError):
            NumEnum(
                ('A', 3), ('B', 2), 'C'
            )

    def test_num_enum_invalid_num_should_raise_exception(self):
        with assert_raises(ValueError):
            NumEnum(
                ('A', 'e'), ('B', 2)
            )

    def test_choices_enum_should_return_right_values_and_choices(self):
        choices_num = ChoicesEnum(
            ('A', 'label a'),
            ('B', 'label b'),
        )
        assert_equal(choices_num.A, 'A')
        assert_equal(choices_num.B, 'B')
        assert_equal(list(choices_num.choices), [('A', 'label a'), ('B', 'label b')])
        assert_equal(choices_num.all, ('A', 'B'))
        assert_equal(choices_num.get_label('A'), 'label a')
        assert_equal(choices_num.get_label('B'), 'label b')

    def test_choices_enum_with_distinct_key_and_value_should_return_right_values_and_choices(self):
        choices_num = ChoicesEnum(
            ('A', 'label a', 'c'),
            ('B', 'label b', 'd'),
        )
        assert_equal(choices_num.A, 'c')
        assert_equal(choices_num.B, 'd')
        assert_equal(list(choices_num.choices), [('c', 'label a'), ('d', 'label b')])
        assert_equal(choices_num.all, ('c', 'd'))
        assert_equal(choices_num.get_label('c'), 'label a')
        assert_equal(choices_num.get_label('d'), 'label b')

    def test_choices_num_enum_should_return_right_values_and_choices(self):
        choices_num = ChoicesNumEnum(
            ('A', 'label a'),
            ('B', 'label b'),
        )
        assert_equal(choices_num.A, 1)
        assert_equal(choices_num.B, 2)
        assert_equal(list(choices_num.choices), [(1, 'label a'), (2, 'label b')])
        assert_equal(tuple(choices_num.all), (1, 2))
        assert_equal(choices_num.get_label(1), 'label a')
        assert_equal(choices_num.get_label(2), 'label b')
        assert_raises(AttributeError, choices_num.get_label, 3)

    def test_choices_num_enum_with_defined_ids_should_return_right_values_and_choices(self):
        choices_num = ChoicesNumEnum(
            ('A', 'label a', 4),
            ('B', 'label b', 2),
            ('C', 'label c')
        )
        assert_equal(choices_num.A, 4)
        assert_equal(choices_num.B, 2)
        assert_equal(choices_num.C, 3)
        assert_equal(list(choices_num.choices), [(4, 'label a'), (2, 'label b'), (3, 'label c')])
        assert_equal(choices_num.all, (4, 2, 3))
        assert_equal(choices_num.get_label(4), 'label a')
        assert_equal(choices_num.get_label(2), 'label b')
        assert_equal(choices_num.get_label(3), 'label c')

    def test_choices_num_enum_same_numbers_should_raise_exception(self):
        with assert_raises(ValueError):
            ChoicesNumEnum(
                ('A', 'label a', 1), ('B', 'label b', 1)
            )

    def test_choices_num_enum_same_generated_numbers_should_raise_exception(self):
        with assert_raises(ValueError):
            ChoicesNumEnum(
                ('A', 'label a', 3), ('B', 'label b', 2), ('C', 'label c')
            )

    def test_choices_num_enum_invalid_num_raise_exception(self):
        with assert_raises(ValueError):
            ChoicesNumEnum(
                ('A', 'label a', 'e'), ('B', 'label b', 2)
            )

    def test_enum_key_should_have_right_format(self):
        with assert_raises(ValueError):
            Enum(
                1, 2
            )
        with assert_raises(ValueError):
            Enum(
                '1A', 'B'
            )
        with assert_raises(ValueError):
            Enum(
                'A-B', 'B'
            )
        Enum(
            'A_B_3', 'B'
        )

    def test_ordered_set_should_keep_order(self):
        ordered_set = OrderedSet(4, 5, 3)
        assert_equal(ordered_set, [4, 5, 3])
        ordered_set.add(8)
        assert_equal(ordered_set, [4, 5, 3, 8])
        assert_equal(ordered_set.pop(), 8)
        assert_equal(ordered_set.pop(False), 4)
        assert_equal(ordered_set, [5, 3])
        ordered_set |= OrderedSet(9, 10)
        assert_equal(ordered_set, [5, 3, 9, 10])
        assert_equal(ordered_set, {5, 3, 9, 10})
        ordered_set |= OrderedSet(5, 9)
        assert_equal(ordered_set, {5, 3, 9, 10})
        ordered_set.discard(9)
        assert_equal(ordered_set, {5, 3, 10})
        ordered_set.add(5)
        ordered_set.add(9)
        assert_equal(ordered_set, {5, 3, 10, 9})



from datetime import timedelta

from django.core.exceptions import ValidationError
from django.test import TransactionTestCase
from django.utils import timezone

from chamber.exceptions import PersistenceException
from chamber.models import DynamicChangedFields, Comparator, Unknown

from germanium.tools import assert_equal, assert_false, assert_raises, assert_true  # pylint: disable=E0401

from test_chamber.models import ComparableModel, DiffModel, RelatedSmartModel, TestSmartModel  # pylint: disable=E0401

from .dispatchers import *  # NOQA
from .fields import *  # NOQA
from .humanized_helpers import *  # NOQA


class NameComparator(Comparator):

    def compare(self, a, b):
        return a.name == b.name


class TestProxySmartModel(TestSmartModel):

    def clean_name(self):
        if len(self.name) >= 10:
            raise ValidationError('name must be lower than 10')

    class Meta:
        proxy = True


class TestPreProxySmartModel(TestSmartModel):

    def __init__(self, **kwargs):
        self.name = None
        super(TestPreProxySmartModel, self).__init__(**kwargs)

    def _pre_save(self, *args, **kwargs):
        self.name = 'test pre save'

    def _pre_delete(self, *args, **kwargs):
        self.name = 'test pre delete'

    class Meta:
        proxy = True


class TestPostProxySmartModel(TestSmartModel):

    def __init__(self, **kwargs):
        self.name = None
        super(TestPostProxySmartModel, self).__init__(**kwargs)

    def _post_save(self, *args, **kwargs):
        self.name = 'test post save'

    def _post_delete(self, *args, **kwargs):
        self.name = 'test post delete'

    class Meta:
        proxy = True


class ModelsTestCase(TransactionTestCase):

    def test_smart_model_initial_values_should_be_unknown_for_not_saved_instance(self):
        obj = DiffModel(name='test', datetime=timezone.now(), number=2)
        assert_true(obj.has_changed)
        assert_true(obj.changed_fields)
        assert_equal(set(obj.changed_fields.keys()), {'created_at', 'changed_at', 'id', 'datetime', 'name', 'number'})
        assert_true(obj.is_adding)
        assert_false(obj.is_changing)
        assert_true(all(v is Unknown for v in obj.initial_values.values()))
        assert_true(all(not bool(v) for v in obj.initial_values.values()))

        obj.save()
        assert_false(obj.has_changed)
        assert_false(obj.changed_fields)
        assert_false(obj.is_adding)
        assert_true(obj.is_changing)
        assert_true(all(v is not Unknown for v in obj.initial_values.values()))

        obj = DiffModel.objects.get(pk=obj.pk)
        assert_false(obj.has_changed)
        assert_false(obj.changed_fields)
        assert_false(obj.is_adding)
        assert_true(obj.is_changing)
        assert_true(all(v is not Unknown for v in obj.initial_values.values()))

    def test_smart_model_changed_fields(self):
        obj = TestProxySmartModel.objects.create(name='a')
        changed_fields = DynamicChangedFields(obj)
        assert_equal(len(changed_fields), 0)
        obj.name = 'b'
        assert_equal(len(changed_fields), 1)
        assert_equal(changed_fields['name'].initial, 'a')
        assert_equal(changed_fields['name'].current, 'b')
        static_changed_fields = changed_fields.get_static_changes()
        obj.save()

        # Initial values is not changed
        assert_equal(len(changed_fields), 2)
        assert_equal(len(static_changed_fields), 1)
        assert_equal(set(changed_fields.keys()), {'name', 'changed_at'})
        assert_equal(set(static_changed_fields.keys()), {'name'})
        assert_equal(changed_fields['name'].initial, 'a')
        assert_equal(changed_fields['name'].current, 'b')

        assert_true(changed_fields.has_any_key('name', 'crated_at'))
        assert_false(changed_fields.has_any_key('invalid', 'crated_at'))

        assert_raises(AttributeError, changed_fields.__delitem__, 'name')
        assert_raises(AttributeError, changed_fields.clear)
        assert_raises(AttributeError, changed_fields.pop, 'name')

        obj.name = 'b'

    def test_model_diff(self):
        obj = DiffModel.objects.create(name='test', datetime=timezone.now(), number=2)
        assert_false(obj.has_changed)
        obj.name = 'test2'
        assert_true(obj.has_changed)
        assert_equal(set(obj.changed_fields.keys()), {'name'})
        assert_equal((obj.changed_fields['name'].initial, obj.changed_fields['name'].current), ('test', 'test2'))

        obj.name = 'test'
        assert_false(obj.has_changed)
        assert_false(obj.changed_fields)

        obj.name = 'test2'
        obj.number = 3
        obj.datetime = obj.datetime + timedelta(days=2)
        assert_true(obj.has_changed)
        assert_equal(set(obj.changed_fields.keys()), {'name', 'number', 'datetime'})

        obj.save()
        assert_false(obj.has_changed)
        assert_false(obj.changed_fields)

    def test_comparator(self):
        obj1 = ComparableModel.objects.create(name='test')
        obj2 = ComparableModel.objects.create(name='test')
        obj3 = ComparableModel.objects.create(name='test2')
        comparator = NameComparator()

        assert_true(obj1.equals(obj2, comparator))
        assert_true(obj2.equals(obj1, comparator))

        assert_false(obj1.equals(obj3, comparator))
        assert_false(obj3.equals(obj1, comparator))

    def test_should_raise_exception_due_calling_default_comparator(self):
        obj1 = ComparableModel.objects.create(name='test')
        obj2 = ComparableModel.objects.create(name='test')

        assert_raises(NotImplementedError, obj1.equals, obj2, Comparator())

    def test_smart_model_clean_pre_save(self):
        assert_raises(PersistenceException, TestProxySmartModel.objects.create, name=10 * 'a')
        obj = TestProxySmartModel.objects.create(name=9 * 'a')
        obj.name = 11 * 'a'
        assert_raises(PersistenceException, obj.save)
        assert_equal(len(TestProxySmartModel.objects.get(pk=obj.pk).name), 9)
        obj.save(is_cleaned_pre_save=False)
        assert_equal(len(TestProxySmartModel.objects.get(pk=obj.pk).name), 11)

    def test_smart_model_clean_post_save(self):
        class PostSaveTestProxySmartModel(TestProxySmartModel):
            class Meta:
                proxy = True
                verbose_name = 'testmodel'
                verbose_name_plural = 'testmodels'

            class SmartMeta:
                is_cleaned_pre_save = False
                is_cleaned_post_save = True

        assert_false(PostSaveTestProxySmartModel.objects.filter(name=10 * 'a').exists())
        assert_raises(PersistenceException, PostSaveTestProxySmartModel.objects.create, name=10 * 'a')
        assert_true(PostSaveTestProxySmartModel.objects.filter(name=10 * 'a').exists())
        obj = PostSaveTestProxySmartModel.objects.create(name=9 * 'a')
        obj.name = 11 * 'a'
        assert_raises(PersistenceException, obj.save)
        assert_equal(len(PostSaveTestProxySmartModel.objects.get(pk=obj.pk).name), 11)
        obj.name = 12 * 'a'
        obj.save(is_cleaned_post_save=False)
        assert_equal(len(PostSaveTestProxySmartModel.objects.get(pk=obj.pk).name), 12)

    def test_smart_model_clean_atomic_post_save(self):
        class AtomicPostSaveTestProxySmartModel(TestProxySmartModel):
            class Meta:
                proxy = True
                verbose_name = 'testmodel'
                verbose_name_plural = 'testmodels'

            class SmartMeta:
                is_cleaned_pre_save = False
                is_cleaned_post_save = True
                is_save_atomic = True

        assert_false(AtomicPostSaveTestProxySmartModel.objects.filter(name=10 * 'a').exists())
        assert_raises(PersistenceException, AtomicPostSaveTestProxySmartModel.objects.create, name=10 * 'a')
        assert_false(AtomicPostSaveTestProxySmartModel.objects.filter(name=10 * 'a').exists())
        obj = AtomicPostSaveTestProxySmartModel.objects.create(name=9 * 'a')
        obj.name = 11 * 'a'
        assert_raises(PersistenceException, obj.save)
        assert_equal(len(AtomicPostSaveTestProxySmartModel.objects.get(pk=obj.pk).name), 9)
        obj.name = 12 * 'a'
        obj.save(is_cleaned_post_save=False)
        assert_equal(len(AtomicPostSaveTestProxySmartModel.objects.get(pk=obj.pk).name), 12)

    def test_smart_model_clean_pre_delete(self):
        class PreDeleteTestProxySmartModel(TestProxySmartModel):
            class Meta:
                proxy = True
                verbose_name = 'testmodel'
                verbose_name_plural = 'testmodels'

            class SmartMeta:
                is_cleaned_pre_save = False
                is_cleaned_pre_delete = True

        obj = PreDeleteTestProxySmartModel.objects.create(name=10 * 'a')
        obj_pk = obj.pk
        assert_raises(PersistenceException, obj.delete)
        assert_true(PreDeleteTestProxySmartModel.objects.filter(pk=obj_pk).exists())

        obj = PreDeleteTestProxySmartModel.objects.create(name=10 * 'a')
        obj_pk = obj.pk
        obj.delete(is_cleaned_pre_delete=False)
        assert_false(PreDeleteTestProxySmartModel.objects.filter(pk=obj_pk).exists())

    def test_smart_model_clean_post_delete(self):
        class PostDeleteTestProxySmartModel(TestProxySmartModel):
            class Meta:
                proxy = True
                verbose_name = 'testmodel'
                verbose_name_plural = 'testmodels'

            class SmartMeta:
                is_cleaned_pre_save = False
                is_cleaned_post_delete = True

        obj = PostDeleteTestProxySmartModel.objects.create(name=10 * 'a')
        obj_pk = obj.pk
        assert_raises(PersistenceException, obj.delete)
        assert_false(PostDeleteTestProxySmartModel.objects.filter(pk=obj_pk).exists())

        obj = PostDeleteTestProxySmartModel.objects.create(name=10 * 'a')
        obj_pk = obj.pk
        obj.delete(is_cleaned_post_delete=False)
        assert_false(PostDeleteTestProxySmartModel.objects.filter(pk=obj_pk).exists())

    def test_smart_model_clean_atomic_post_delete(self):
        class AtomicPostDeleteTestProxySmartModel(TestProxySmartModel):
            class Meta:
                proxy = True
                verbose_name = 'testmodel'
                verbose_name_plural = 'testmodels'

            class SmartMeta:
                is_cleaned_pre_save = False
                is_cleaned_post_delete = True
                is_delete_atomic = True

        obj = AtomicPostDeleteTestProxySmartModel.objects.create(name=10 * 'a')
        obj_pk = obj.pk
        assert_raises(PersistenceException, obj.delete)
        assert_true(AtomicPostDeleteTestProxySmartModel.objects.filter(pk=obj_pk).exists())

        obj = AtomicPostDeleteTestProxySmartModel.objects.create(name=10 * 'a')
        obj_pk = obj.pk
        obj.delete(is_cleaned_post_delete=False)
        assert_false(AtomicPostDeleteTestProxySmartModel.objects.filter(pk=obj_pk).exists())

    def test_smart_model_pre_save(self):
        obj = TestPreProxySmartModel.objects.create()
        assert_equal(obj.name, 'test pre save')
        obj.name = 10 * 'a'
        obj.save()
        assert_equal(obj.name, 'test pre save')
        assert_true(TestPreProxySmartModel.objects.filter(name='test pre save').exists())

    def test_smart_model_pre_delete(self):
        obj = TestPreProxySmartModel.objects.create()
        assert_equal(obj.name, 'test pre save')
        obj.delete()
        assert_equal(obj.name, 'test pre delete')

    def test_smart_model_post_save(self):
        assert_raises(PersistenceException, TestPostProxySmartModel.objects.create)
        obj = TestPostProxySmartModel.objects.create(name=10 * 'a')
        assert_equal(obj.name, 'test post save')
        assert_false(TestPreProxySmartModel.objects.filter(name='test post save').exists())
        assert_true(TestPreProxySmartModel.objects.filter(name=10 * 'a').exists())
        obj.save()
        assert_true(TestPreProxySmartModel.objects.filter(name='test post save').exists())
        obj.name = 10 * 'a'
        obj.save()
        assert_equal(obj.name, 'test post save')
        assert_false(TestPreProxySmartModel.objects.filter(name='test post save').exists())

    def test_smart_model_post_delete(self):
        obj = TestPostProxySmartModel.objects.create(name=10 * 'a')
        assert_equal(obj.name, 'test post save')
        obj.delete()
        assert_equal(obj.name, 'test post delete')

    def test_smart_queryset_fast_distinct(self):
        t = TestSmartModel.objects.create(name='name')
        RelatedSmartModel.objects.create(test_smart_model=t)
        RelatedSmartModel.objects.create(test_smart_model=t)
        qs = TestSmartModel.objects.filter(test_smart_models__test_smart_model=t)
        assert_equal(qs.count(), 2)
        assert_equal(tuple(qs.values_list('pk', flat=True)), (t.pk, t.pk))
        assert_equal(qs.fast_distinct().count(), 1)



from setuptools import setup, find_packages

from chamber.version import get_version


setup(
    name='django-chamber',
    version=get_version(),
    description='Utilities library meant as a complement to django-is-core.',
    author='Lubos Matl, Oskar Hollmann',
    author_email='matllubos@gmail.com, oskar@hollmann.me',
    url='http://github.com/druids/django-chamber',
    packages=find_packages(),
    package_dir={'chamber': 'chamber'},
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Framework :: Django',
    ],
    install_requires=[
        'Django>=1.8',
        'Unidecode>=0.04.17',
        'pyprind>=2.11.0',
        'six>=1.10.0',
        'filemagic>=1.6',
    ],
    extras_require={
        'boto3storage': ['django-storages<2.0', 'boto3'],
    },
)



from django.test import TestCase

from chamber.utils.transaction import (
    on_success, transaction_signals, OnSuccessHandler, OneTimeOnSuccessHandler
)

from germanium.tools import assert_equal  # pylint: disable=E0401


def add_number(numbers_list, number):
    numbers_list.append(number)


class TransactionsTestCase(TestCase):

    def test_on_success_is_called_after_successful_pass(self):
        numbers_list = []

        with transaction_signals():
            on_success(lambda: add_number(numbers_list, 0))

            assert_equal(len(numbers_list), 0)

        assert_equal(len(numbers_list), 1)

    def test_on_success_is_not_called_after_not_successful_pass(self):
        numbers_list = []

        try:
            with transaction_signals():
                on_success(lambda: add_number(numbers_list, 0))

                assert_equal(len(numbers_list), 0)
                raise Exception()
        except Exception:
            pass

        assert_equal(len(numbers_list), 0)

    def test_on_success_inheritance(self):
        numbers_list = []

        with transaction_signals():
            with transaction_signals():
                on_success(lambda: add_number(numbers_list, 0))

                assert_equal(len(numbers_list), 0)
            assert_equal(len(numbers_list), 0)
            try:
                with transaction_signals():
                    on_success(lambda: add_number(numbers_list, 1))

                    assert_equal(len(numbers_list), 0)
                    raise Exception()
            except Exception:
                pass
            on_success(lambda: add_number(numbers_list, 2))
            assert_equal(len(numbers_list), 0)

        assert_equal(numbers_list, [0,2])

    def test_on_success_handler(self):
        numbers_list = []

        class AddNumberOnSuccessHandler(OnSuccessHandler):

            def handle(self, numbers_list, number):
                numbers_list.append(number)

        with transaction_signals():
            AddNumberOnSuccessHandler(numbers_list=numbers_list, number=1)
            AddNumberOnSuccessHandler(numbers_list=numbers_list, number=2)

            assert_equal(len(numbers_list), 0)

        assert_equal(numbers_list, [1, 2])

    def test_on_success_one_time_handler(self):
        numbers_list = []

        class AddNumberOneTimeOnSuccessHandler(OneTimeOnSuccessHandler):

            def handle(self, kwargs_list):
                kwargs_list[-1]['numbers_list'].append(3)

        with transaction_signals():
            for i in range(5):
                AddNumberOneTimeOnSuccessHandler(numbers_list=numbers_list, number=i)

            assert_equal(len(numbers_list), 0)

        assert_equal(numbers_list, [3])

    def test_on_success_one_time_handler_inheritance(self):
        numbers_list = []

        class AddNumberOneTimeOnSuccessHandler(OneTimeOnSuccessHandler):

            def handle(self, kwargs_list):
                for kwargs in kwargs_list:
                    kwargs_list[-1]['numbers_list'].append(kwargs['number'])

        with transaction_signals():
            AddNumberOneTimeOnSuccessHandler(numbers_list=numbers_list, number=1)
            try:
                with transaction_signals():
                    AddNumberOneTimeOnSuccessHandler(numbers_list=numbers_list, number=2)

                    assert_equal(len(numbers_list), 0)
                    raise Exception()
            except Exception:
                pass
            with transaction_signals():
                AddNumberOneTimeOnSuccessHandler(numbers_list=numbers_list, number=3)

                assert_equal(len(numbers_list), 0)
            assert_equal(len(numbers_list), 0)

        assert_equal(numbers_list, [1, 3])



from collections import defaultdict

from django.core.exceptions import ImproperlyConfigured

from chamber.utils.transaction import (
    on_success, OnSuccessHandler, OneTimeOnSuccessHandler, InstanceOneTimeOnSuccessHandler
)

from .signals import dispatcher_post_save


class BaseDispatcher:
    """
    Base dispatcher class that can be subclassed to call a handler based on a change in some a SmartModel.
    If you subclass, be sure the __call__ method does not change signature.
    """

    signal = None

    def _validate_init_params(self):
        if not callable(self.handler):
            raise ImproperlyConfigured('Registered handler must be a callable.')

    def __init__(self, handler, signal=None):
        self.handler = handler
        self._validate_init_params()
        self._connected = defaultdict(list)
        self._signal = signal if signal is not None else self.signal

    def connect(self, sender):
        self._signal.connect(self, sender=sender)

    def __call__(self, instance, **kwargs):
        """
        `instance` ... instance of the SmartModel where the handler is being called
        Some dispatchers require additional params to evaluate the handler can be dispatched,
        these are hidden in args and kwargs.
        """
        if self._can_dispatch(instance, **kwargs):
            self.handler(instance=instance, **kwargs)

    def _can_dispatch(self, instance, *args, **kwargs):
        raise NotImplementedError


class PropertyDispatcher(BaseDispatcher):
    """
    Use this class to register a handler to dispatch during save if the given property evaluates to True.
    """

    def _validate_init_params(self):
        """
        No validation is done as it would require to pass the whole model to the dispatcher.
        If the property is not defined, a clear error is shown at runtime.
        """
        pass

    def __init__(self, handler, property_name, signal=None):
        self.property_name = property_name
        super().__init__(handler, signal)

    def _can_dispatch(self, instance, **kwargs):
        return getattr(instance, self.property_name)


class CreatedDispatcher(BaseDispatcher):
    """
    Calls registered handler if and only if an instance of the model is being created.
    """

    def _can_dispatch(self, instance, change, **kwargs):
        return not change


class StateDispatcher(BaseDispatcher):
    """
    Use this class to register a handler for transition of a model to a certain state.
    """

    def _validate_init_params(self):
        super()._validate_init_params()
        if self.field_value not in {value for value, _ in self.enum.choices}:
            raise ImproperlyConfigured('Enum of FieldDispatcher does not contain {}.'.format(self.field_value))

    def __init__(self, handler, enum, field, field_value, signal=None):
        self.enum = enum
        self.field = field
        self.field_value = field_value
        super().__init__(handler, signal=signal)

    def _can_dispatch(self, instance, change, changed_fields, *args, **kwargs):
        return (
            self.field.get_attname() in changed_fields and
            getattr(instance, self.field.get_attname()) == self.field_value
        )



from django.db.models import Model
from django.db.models.fields import Field


class OptionsLazy:

    def __init__(self, name, klass):
        self.name = name
        self.klass = klass

    def __get__(self, instance=None, owner=None):
        return self.klass(owner)


class OptionsBase(type):
    def __new__(cls, *args, **kwargs):
        new_class = super().__new__(cls, *args, **kwargs)
        if new_class.model_class and new_class.meta_name:
            setattr(new_class.model_class, new_class.meta_name, OptionsLazy(new_class.meta_name, new_class))
        return new_class


class Options(metaclass=OptionsBase):

    meta_class_name = None
    meta_name = None
    attributes = None
    model_class = None

    def __init__(self, model):
        self.model = model

        for key, default_value in self._get_attributes(model).items():
            setattr(self, key, self._getattr(key, default_value))

    def _get_attributes(self, model):
        return self.attributes

    def _getattr(self, name, default_value):
        meta_models = [b for b in self.model.__mro__ if issubclass(b, Model)]
        for model in meta_models:
            meta = getattr(model, self.meta_class_name, None)
            if meta:
                value = getattr(meta, name, None)
                if value is not None:
                    return value
        return default_value


def field_init(self, *args, **kwargs):
    """
    Patches a Django Field's `__init__` method for easier usage of optional `kwargs`. It defines a `humanized` attribute
    on a field for better display of its value.
    """
    humanize_func = kwargs.pop('humanized', None)
    if humanize_func:
        def humanize(val, inst, *args, **kwargs):
            return humanize_func(val, inst, field=self, *args, **kwargs)
        self.humanized = humanize
    else:
        self.humanized = self.default_humanized
    getattr(self, '_init_chamber_patch_')(*args, **kwargs)


Field.default_humanized = None
Field._init_chamber_patch_ = Field.__init__  # pylint: disable=W0212
Field.__init__ = field_init



from django.contrib.auth.backends import ModelBackend as OriginModelBackend
from django.contrib.auth.models import Permission

from chamber.multidomains.domain import get_user_class


class GetUserMixin:

    def get_user(self, user_id):
        UserModel = get_user_class()
        try:
            return UserModel._default_manager.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None


class ModelBackend(GetUserMixin, OriginModelBackend):
    """
    Authenticates against settings.AUTH_USER_MODEL.
    """

    def authenticate(self, username=None, password=None, **kwargs):
        UserModel = get_user_class()
        if username is None:
            username = kwargs.get(UserModel.USERNAME_FIELD)
        try:
            user = UserModel._default_manager.get_by_natural_key(username)  # pylint: disable=W0212
            if user.check_password(password):
                return user
        except UserModel.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a non-existing user (#20760).
            UserModel().set_password(password)

    def get_group_permissions(self, user_obj, obj=None):
        """
        Returns a set of permission strings that this user has through his/her
        groups.
        """
        if user_obj.is_anonymous() or obj is not None:
            return set()
        if not hasattr(user_obj, '_group_perm_cache'):
            if user_obj.is_superuser:
                perms = Permission.objects.all()
            else:
                user_groups_field = get_user_class()._meta.get_field('groups')  # pylint: disable=W0212
                user_groups_query = 'group__%s' % user_groups_field.related_query_name()
                perms = Permission.objects.filter(**{user_groups_query: user_obj})
            perms = perms.values_list('content_type__app_label', 'codename').order_by()
            user_obj._group_perm_cache = set(["%s.%s" % (ct, name) for ct, name in perms])  # pylint: disable=W0212
        return user_obj._group_perm_cache  # pylint: disable=W0212



from django.core.files.base import ContentFile

from storages.backends.s3boto3 import S3Boto3Storage


__all__ = (
    'S3Storage',
    'force_bytes_content',
)


def force_bytes_content(content, blocksize=1024):
    '''Returns a tuple of content (file-like object) and bool indicating wheter the content has been casted or not'''
    block = content.read(blocksize)
    content.seek(0)

    if not isinstance(block, bytes):
        _content = bytes(
            content.read(),
            'utf-8' if not hasattr(content, 'encoding') or content.encoding is None else content.encoding,
        )
        return ContentFile(_content), True
    return content, False


class S3Storage(S3Boto3Storage):

    def save(self, name, content, max_length=None):
        content, _ = force_bytes_content(content)
        return super().save(name, content, max_length)
